<?php
namespace Inti\Gath\Model\Repository;

use Inti\Gath\Api\Endpoint\GetProductAttributes;
use Inti\Gath\Api\Endpoint\GetProductAttributesByCategory;
use Inti\Gath\Api\Query\GetProductAttributesByCategoryQuery;
use Inti\Gath\Model\Entity\Attribute;
use Inti\Gath\Model\Entity\Category;
use Inti\Gath\Model\Hydrator\Strategy\CategoryIdStrategy;

/**
 * Class AttributeRepository
 *
 * @package Inti\Gath\Model\Repository
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-gath
 */
final class AttributeRepository extends AbstractRepository
{
    /** @var Attribute[] */
    private $all;

    /**
     * @return Attribute[]
     */
    public function getAll(): array
    {
        if ($this->all === null) {
            /** @var GetProductAttributes $endpoint */
            $endpoint = $this->endpoints->get(GetProductAttributes::class);
            $result = $endpoint->call();

            foreach ($result->Results as $attributeRecord) {
                $data = $attributeRecord->toArray();
                $this->all[] = $this->hydrator->hydrate($data, new Attribute());
            }
        }

        return $this->all ?? [];
    }

    /**
     * @param Category $category
     * @return Attribute[]
     */
    public function getByCategoryAndItsChildren(Category $category): array
    {
        $categoryIds = $this->getAllCategoryIds($category);
        return $this->getByCategories($categoryIds);
    }

    /**
     * @param int[] $categoryIds
     * @return Attribute[]
     */
    public function getByCategories(array $categoryIds): array
    {
        $categoryIdStrategy = new CategoryIdStrategy();

        $query = new GetProductAttributesByCategoryQuery([
            'Categories' => array_map([$categoryIdStrategy, 'extract'], $categoryIds),
        ]);

        /** @var GetProductAttributesByCategory $endpoint */
        $endpoint = $this->endpoints->get(GetProductAttributesByCategory::class);
        $result = $endpoint->call($query);

        foreach ($result->Results as $attributeRecord) {
            $data = $attributeRecord->toArray();
            $attributes[] = $this->hydrator->hydrate($data, new Attribute());
        }

        return $attributes ?? [];
    }

    /**
     * @param int $id
     * @return Attribute|null
     */
    public function find(int $id): ?Attribute
    {
        foreach ($this->getAll() as $attribute) {
            if ($attribute->getId() === $id) {
                return $attribute;
            }
        }

        return null;
    }

    /**
     * @param string $slug
     * @return Attribute|null
     */
    public function findBySlug(string $slug): ?Attribute
    {
        foreach ($this->getAll() as $attribute) {
            if ($attribute->getSlug() === $slug) {
                return $attribute;
            }
        }

        return null;
    }

    /**
     * @param Category $category
     * @return int[]
     */
    private function getAllCategoryIds(Category $category): array
    {
        $categoryIds[] = $category->getId();

        foreach ($category->getChildren() as $child) {
            $categoryIds = array_merge($categoryIds, $this->getAllCategoryIds($child));
        }

        return $categoryIds;
    }
}
