<?php
namespace Inti\Gath\Model\Repository;

use Inti\Gath\Api\Endpoint\GetProductCategories;
use Inti\Gath\Api\Query\GetProductCategoriesQuery;
use Inti\Gath\Api\Record\ProductCategoryRecord;
use Inti\Gath\Model\Entity\Category;

/**
 * Class CategoryRepository
 *
 * @package Inti\Gath\Model\Repository
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-gath
 */
final class CategoryRepository extends AbstractRepository
{
    /** @var Category[] */
    private $all;

    /**
     * @return Category[]
     */
    public function getAll(): array
    {
        if ($this->all === null) {
            /** @var GetProductCategories $endpoint */
            $endpoint = $this->endpoints->get(GetProductCategories::class);
            $result = $endpoint->call(new GetProductCategoriesQuery());
            $categoryRecords = $this->flattenHierarchy($result->Results ?? []);

            foreach ($categoryRecords as $categoryRecord) {
                $data = $categoryRecord->toArray();
                $this->all[] = $this->hydrator->hydrate($data, new Category());
            }
        }

        return $this->all ?? [];
    }

    /**
     * @param array $ids
     * @return Category[]
     */
    public function getByIds(array $ids): array
    {
        return array_values(array_filter($this->getAll(), function (Category $category) use ($ids) {
            return in_array($category->getId(), $ids);
        }));
    }

    /**
     * @param int $id
     * @return Category|null
     */
    public function getById(int $id): ?Category
    {
        $categories = $this->getByIds([$id]);
        return $categories[0] ?? null;
    }

    /**
     * @param string[] $slugs
     * @return Category[]
     */
    public function getBySlugs(array $slugs): array
    {
        return array_values(array_filter($this->getAll(), function(Category $category) use ($slugs) {
            return in_array($category->getSlug(), $slugs);
        }));
    }

    /**
     * @param int $parentId
     * @return Category[]
     */
    public function getByParentId(int $parentId): array
    {
        return array_filter($this->getAll(), function (Category $category) use ($parentId) {
            return $category->getParentId() === $parentId;
        });
    }

    /**
     * @param ProductCategoryRecord[] $hierarchy
     * @return ProductCategoryRecord[]
     */
    private function flattenHierarchy(array $hierarchy)
    {
        $flattened = [];

        foreach ($hierarchy as $category) {
            $flattened[] = $category;
            $children = $this->flattenHierarchy($category->Children ?? []);
            $flattened = array_merge($flattened, $children);
        }

        return $flattened;
    }
}
