<?php
namespace Inti\Gath\Model\Repository;

use Inti\Gath\Api\Endpoint\GetFeaturedProducts;
use Inti\Gath\Api\Endpoint\GetProductsByAttributes;
use Inti\Gath\Api\Endpoint\GetProductsByCategory;
use Inti\Gath\Api\Endpoint\GetProductsByIds;
use Inti\Gath\Api\Endpoint\Service\EndpointManager;
use Inti\Gath\Api\Query\AttributeValue;
use Inti\Gath\Api\Query\GetFeaturedProductsQuery;
use Inti\Gath\Api\Query\GetProductsByAttributesQuery;
use Inti\Gath\Api\Query\GetProductsByCategoryQuery;
use Inti\Gath\Api\Query\GetProductsByIdsQuery;
use Inti\Gath\Api\Result\GetProductsResult;
use Inti\Gath\Model\Entity\Attribute;
use Inti\Gath\Model\Entity\Category;
use Inti\Gath\Model\Entity\Product;
use Inti\Gath\Model\Hydrator\Strategy\CategoryIdStrategy;
use Inti\Gath\Model\Hydrator\Strategy\ProductIdStrategy;
use Inti\Gath\Model\Hydrator\Strategy\TireIdStrategy;
use Laminas\Hydrator\HydratorInterface;

/**
 * Class ProductRepository
 *
 * @package Inti\Gath\Model\Repository
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-gath
 */
final class ProductRepository extends AbstractRepository
{
    const SORT_DATE_DESC = 1;
    const SORT_PRICE_ASC = 2;
    const SORT_PRICE_DESC = 3;
    const SORT_STOCK_DESC = 4;
    const SORT_NAME_DESC = 5;
    const SORT_NAME_ASC = 6;

    /** @var AttributeRepository */
    private $attributes;
    /** @var CategoryRepository */
    private $categories;
    /** @var int */
    private $totalProducts = 0;

    /**
     * ProductRepository constructor.
     *
     * @param EndpointManager $endpoints
     * @param HydratorInterface $hydrator
     * @param AttributeRepository $attributes
     * @param CategoryRepository $categories
     */
    public function __construct(
        EndpointManager $endpoints,
        HydratorInterface $hydrator,
        AttributeRepository $attributes,
        CategoryRepository $categories
    ) {
        parent::__construct($endpoints, $hydrator);
        $this->attributes = $attributes;
        $this->categories = $categories;
    }

    /**
     * @inheritDoc
     */
    public function getAll(int $limit = 1000, int $offset = 0, int $sort = self::SORT_NAME_ASC): array
    {
        $query = $this->getProductsByAllAttributesQuery($limit, $offset, $sort);

        /** @var GetProductsByAttributes $endpoint */
        $endpoint = $this->endpoints->get(GetProductsByAttributes::class);
        $result = $endpoint->call($query);

        return $this->hydrateResult($result);
    }

    /**
     * @param string $slug
     * @param int $limit
     * @param int $offset
     * @param int $sort
     * @return Product[]
     */
    public function getByCategorySlug(string $slug, int $limit = 1000, int $offset = 0, int $sort = self::SORT_NAME_ASC): array
    {
        $category = $this->categories->getBySlugs([$slug])[0] ?? null;

        if (!$category) {
            return [];
        }

        /*if (count($category->getChildren()) > 0) {
            $query = $this->getProductsByAllAttributesQuery($limit, $offset, $sort, $category);
            $categoryIdStrategy = new CategoryIdStrategy();
            $query->Categories = array_map([$categoryIdStrategy, 'extract'], $this->getAllCategoryIds($category));

            /** @var GetProductsByAttributes $endpoint */
            //$endpoint = $this->endpoints->get(GetProductsByAttributes::class);
        //} else {
            $query = new GetProductsByCategoryQuery([
                'CategorySlug' => $slug,
                'Limit' => $limit,
                'Offset' => $offset,
                'OrderBy' => $sort,
            ]);

            /** @var GetProductsByCategory $endpoint */
            $endpoint = $this->endpoints->get(GetProductsByCategory::class);
        //}

        $result = $endpoint->call($query);
        return $this->hydrateResult($result);
    }

    /**
     * @param array $ids
     * @return Product[]
     */
    public function getByIds(array $ids): array
    {
        $query = new GetProductsByIdsQuery();
        $tireIdStrategy = new TireIdStrategy();
        $productIdStrategy = new ProductIdStrategy();

        foreach ($ids as $id) {
            if ($id > TireIdStrategy::ID_OFFSET) {
                $TireIds[] = $tireIdStrategy->extract($id);
            } else if ($id > ProductIdStrategy::ID_OFFSET) {
                $Ids[] = $productIdStrategy->extract($id);
            }
        }

        $query->Ids = $Ids ?? null;
        $query->TireIds = $TireIds ?? null;

        /** @var GetProductsByIds $endpoint */
        $endpoint = $this->endpoints->get(GetProductsByIds::class);
        $result = $endpoint->call($query);

        return $this->hydrateResult($result);
    }

    public function getFeaturedProducts(): array
    {
        $query = new GetFeaturedProductsQuery();
        $query->OrderBy = 1;
        $query->Limit = 1000;
        $query->Offset = 0;

        /** @var GetFeaturedProducts $endpoint */
        $endpoint = $this->endpoints->get(GetFeaturedProducts::class);
        $result = $endpoint->call($query);

        return $this->hydrateResult($result);
    }

    /**
     * @param array $attributes
     * @param int|null $categoryId
     * @param int $limit
     * @param int $offset
     * @param int $sort
     * @return Product[]
     */
    public function getByAttributes(array $attributes, int $categoryId = null, int $limit = 1000, int $offset = 0, int $sort = self::SORT_NAME_ASC): array
    {
        foreach ($attributes as $attributeSlug => $valueSlugs) {
            $attribute = $this->attributes->findBySlug($attributeSlug);

            foreach ($valueSlugs as $valueSlug) {
                foreach ($attribute->getAllValues() as $unsanitizedValue) {
                    if ($valueSlug === sanitize_title($unsanitizedValue)) {
                        $values[] = $unsanitizedValue;
                    }
                }
            }

            if (isset($values)) {
                $attributeQuery[] = new AttributeValue($attribute->getId(), $values);
            }
        }

        if ($categoryId) {
            $category = $this->categories->getById($categoryId);
            $categoryIds = $this->getAllCategoryIds($category);

            foreach ($categoryIds as $i => $categoryId) {
                if ($categoryId > 1e6) {
                    $categoryIds[$i] = intval($categoryId - 1e6);
                }
            }
        }

        $query = new GetProductsByAttributesQuery([
            'Attributes' => $attributeQuery ?? [],
            'Categories' => $categoryIds ?? null,
            'Limit' => $limit,
            'Offset' => $offset,
            'OrderBy' => $sort,
        ]);

        /** @var GetProductsByAttributes $endpoint */
        $endpoint = $this->endpoints->get(GetProductsByAttributes::class);
        $result = $endpoint->call($query);

        return $this->hydrateResult($result);
    }

    /**
     * @param int $id
     * @return Product|null
     */
    public function getById(int $id): ?Product
    {
        $products = $this->getByIds([$id]);
        return $products[0] ?? null;
    }

    /**
     * @return int
     */
    public function getTotalProducts(): int
    {
        return $this->totalProducts;
    }

    /**
     * @param GetProductsResult $result
     * @return Product[]
     */
    private function hydrateResult(GetProductsResult $result): array
    {
        foreach ($result->Results as $productRecord) {
            $data = $productRecord->toArray();
            $products[] = $this->hydrator->hydrate($data, new Product());
        }

        $this->totalProducts = $result->RecordCount ?? 0;
        return $products ?? [];
    }

    /**
     * @param int $limit
     * @param int $offset
     * @param int $sort
     * @param Category|null $category
     * @return GetProductsByAttributesQuery
     */
    private function getProductsByAllAttributesQuery(int $limit, int $offset, int $sort, Category $category = null): GetProductsByAttributesQuery
    {
        if ($category instanceof Category) {
            $categoryIds = $this->getAllCategoryIds($category);
            $attributes = $this->attributes->getByCategories($categoryIds);
        } else {
            $attributes = $this->attributes->getAll();
        }

        return $query = new GetProductsByAttributesQuery([
            'Attributes' => array_map(function (Attribute $attribute) {
                return new AttributeValue($attribute->getId(), $attribute->getAllValues());
            }, $attributes),
            'Limit' => $limit,
            'Offset' => $offset,
            'OrderBy' => $sort,
        ]);
    }

    private function getAllCategoryIds(Category $category)
    {
        $ids[] = $category->getId();

        foreach ($category->getChildren() as $child) {
            $ids = array_merge($ids, $this->getAllCategoryIds($child));
        }

        return $ids;
    }
}
