<?php
namespace Inti\Gath\Wp\Plugin;

use Inti\Gath\Model\Entity\Attribute;
use Inti\Gath\Model\Hydrator\Strategy\IdStrategy;
use Inti\Gath\Model\Repository\AttributeRepository;
use Inti\Gath\Model\Repository\CategoryRepository;
use WP_Term;
use WP_Term_Query;

/**
 * Class GathAttributes
 *
 * @package Inti\Gath\Wp\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-gath
 */
final class GathAttributes extends GathTerms
{
    /** @var AttributeRepository */
    private $attributes;
    /** @var CategoryRepository */
    private $categories;

    /**
     * Konstruktor.
     *
     * @param AttributeRepository $attributes
     * @param CategoryRepository $categories
     */
    public function __construct(AttributeRepository $attributes, CategoryRepository $categories)
    {
        $this->attributes = $attributes;
        $this->categories = $categories;
    }

    /**
     * @inheritDoc
     */
    public function init(): void
    {
        parent::init();

        add_action('admin_notices', [$this, 'admin_notices']);
        add_filter('pre_transient_wc_attribute_taxonomies', [$this, 'get_attribute_taxonomies']);
        add_filter('inti/webshop/attributes_for_category', [$this, 'get_attributes_for_category'], 10, 2);
    }

    public function admin_notices()
    {
        global $pagenow;

        if ($pagenow === 'edit.php' && ($_GET['page'] ?? '') === 'product_attributes'): ?>
            <div class="notice notice-warning">
                <p>Atributi se uređuju u Gathu.</p>
            </div>
        <?php endif;
    }

    /**
     * @return object[]
     */
    public function get_attribute_taxonomies(): array
    {
        $attributes = $this->attributes->getAll();
        return $this->hydrateTaxonomies($attributes);
    }

    /**
     * @param $attributes
     * @param int $categoryId
     * @return object[]
     * @noinspection PhpUnusedParameterInspection
     */
    public function get_attributes_for_category($attributes, int $categoryId): array
    {
        $category = $this->categories->getById($categoryId);
        $attributes = $this->attributes->getByCategoryAndItsChildren($category);
        return $this->hydrateTaxonomies($attributes);
    }

    /**
     * @param array $taxonomies
     * @return bool
     */
    protected function is_valid_taxonomy(array $taxonomies): bool
    {
        foreach ($taxonomies as $taxonomy) {
            if (strpos($taxonomy, 'pa_') === 0) {
                return true;
            }
        }

        return false;
    }

    /**
     * @param WP_Term_Query $query
     * @return WP_Term[]
     */
    protected function get_all_terms(WP_Term_Query $query): array
    {
        foreach ($this->get_queried_taxonomies($query) as $taxonomy) {
            if (!preg_match('/^pa_(.+)$/', $taxonomy, $matches)) {
                continue;
            }

            $attributeSlug = $matches[1];
            $attribute = $this->attributes->findBySlug($attributeSlug);
            $attributeValues = $attribute->getAllValues();
            sort($attributeValues);

            foreach ($attributeValues as $attributeValue) {
                $terms[] = $this->load_term($attribute, $attributeValue);
            }
        }

        return $terms ?? [];
    }

    /**
     * @param array $term_ids
     * @return array
     */
    protected function get_terms_by_id(array $term_ids): array
    {
        foreach ($term_ids as $term_id) {
            if (preg_match('/^(\d{1,3})(\d{3})$/', (string) intval($term_id - IdStrategy::ID_OFFSET), $matches)) {
                $attributeId = intval($matches[1]);
                $valueIndex = intval($matches[2]);

                $attribute = $this->attributes->find($attributeId);
                $value = $attribute->getAllValues()[$valueIndex];

                $terms[] = $this->load_term($attribute, $value);
            }
        }

        return $terms ?? [];
    }

    /**
     * @param Attribute $attribute
     * @param string $value
     * @return WP_Term
     */
    private function load_term(Attribute $attribute, string $value): WP_Term
    {
        $valueIndex = array_search($value, $attribute->getAllValues());
        $attributeTermId = intval(sprintf('%03d%03d', $attribute->getId(), $valueIndex) + IdStrategy::ID_OFFSET);
        $attributeTaxonomy = 'pa_' . $attribute->getSlug();

        return new WP_Term((object) [
            'term_id' => $attributeTermId,
            'name' => $value,
            'slug' => sanitize_title($value),
            'taxonomy' => $attributeTaxonomy,
            'description' => '',
            'parent' => 0,
        ]);
    }

    /**
     * @param Attribute[] $attributes
     * @return array<string, object>
     */
    private function hydrateTaxonomies(array $attributes): array
    {
        foreach ($attributes as $attribute) {
            $taxonomies['id:' . $attribute->getId()] = (object) [
                'attribute_id' => (string) $attribute->getId(),
                'attribute_name' => $attribute->getSlug(),
                'attribute_label' => $attribute->getName(),
                'attribute_type' => 'select',
                'attribute_orderby' => 'name',
                'attribute_public' => 0,
                'attribute_values' => $attribute->getAllValues(),
            ];
        }

        return $taxonomies ?? [];
    }
}
