<?php
namespace Inti\Gath\Wp\Plugin;

use Exception;
use Inti\Gath\Model\Entity\Product;
use Inti\Gath\Model\Hydrator\Strategy\IdStrategy;
use Inti\Gath\Model\Hydrator\Strategy\ProductIdStrategy;
use Inti\Gath\Model\Repository\ProductRepository;
use Inti\Gath\Wp\ImageManager;
use Solarium\Client as SolrClient;
use WP_Post;
use WP_Query;

/**
 * Class GathProducts
 *
 * @package Inti\Gath\Wp\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-gath
 */
final class GathProducts extends GathMetadata
{
    /** @var ProductRepository */
    private $repo;
    /** @var ImageManager */
    private $images;
    /** @var SolrClient */
    private $solr;
    /** @var Product[] */
    private $products = [];
    /** @var string */
    protected $metaType = 'post';
    /** @var string[] */
    protected $metaKeys = [
        '_sku',
        'total_sales',
        '_tax_status',
        '_tax_class',
        '_backorders',
        '_low_stock_amount',
        '_sold_individually',
        '_weight',
        '_length',
        '_width',
        '_height',
        '_upsell_ids',
        '_crosssell_ids',
        '_purchase_note',
        '_default_attributes',
        '_virtual',
        '_downloadable',
        '_download_limit',
        '_download_expiry',
        '_thumbnail_id',
        '_wc_average_rating',
        '_wc_rating_count',
        '_wc_review_count',
        '_product_image_gallery',
    ];

    /**
     * GathProducts constructor.
     * @param ProductRepository $repo
     * @param ImageManager $images
     * @param SolrClient $solr
     */
    public function __construct(ProductRepository $repo, ImageManager $images, ?SolrClient $solr)
    {
        $this->repo = $repo;
        $this->images = $images;
        $this->solr = $solr;
    }

    /**
     * @inheritDoc
     * @noinspection PhpUnusedParameterInspection
     */
    public function init(): void
    {
        parent::init();

        add_filter('posts_pre_query', [$this, 'get_posts'], 0, 2);
        add_filter('the_post', [$this, 'setup_post']);
        add_filter('woocommerce_add_to_cart_product_id', [$this, 'wc_add_to_cart_product_id']);
        add_action('init', [$this, 'preload_cart_products']);

        add_filter('woocommerce_get_order_item_classname', function ($classname, $type, $id) {
            if ($product_id = wc_get_order_item_meta($id, '_product_id', true)) {
                get_posts([
                    'post_type' => 'product',
                    'include' => [$product_id],
                ]);
            }
            return $classname;
        }, 10, 3);

        add_action('woocommerce_checkout_order_processed', function ($order_id) {
            $order_type = defined('GATH_WHOLESALE') && GATH_WHOLESALE === true ? 'wholesale' : 'retail';
            update_post_meta($order_id, '_order_type', $order_type);
        });

        add_filter('woocommerce_get_product_terms', [$this, 'get_product_terms'], 10, 3);
    }

    public function get_product_terms($terms, $product_id, $taxonomy)
    {
        if ($taxonomy !== 'product_cat' || count($terms ?? []) < 3 || !isset($this->products[$product_id])) {
            return $terms;
        }

        $product = $this->products[$product_id];
        $categoryIds = $product->getCategoryIds();

        foreach ($categoryIds as &$categoryId) {
            $categoryId += 1e6;
        }

        if (count($categoryIds) === 0) {
            return $terms;
        }

        $terms = get_terms([
            'taxonomy' => 'product_cat',
            'include' => $categoryIds
        ]);

        return $terms;
    }

    /**
     * @param $queryString
     *
     * @return \Solarium\QueryType\Select\Result\Result
     */
    protected function getSearchResults($queryString)
    {
        $searchQuery = $this->solr->createSelect();
        $searchQuery->setQuery($queryString);
        $searchQuery->setRows(10000);

        return $this->solr->select($searchQuery);
    }

    private function product_search(WP_Query $query): WP_Query
    {
        if (!isset($this->solr)) {
            return $query;
        }

        $q = $query->get('s');
        $page = $query->get('paged') ?: 1;

        if (strlen($q) > 0) {
            try {
                $queryString = sprintf('name:"%1$s" OR sku:"%2$s" OR attribute:"%1$s"', $q, strtoupper($q));
                $results = $this->getSearchResults($queryString);

                if ($results->count() === 0) {
                    $queryString = sprintf('name:%1$s~1 OR sku:%1$s OR attribute:%1$s', $q);
                    $results = $this->getSearchResults($queryString);
                }
            } catch (Exception $e) {
                $results = [];
            }

            $productIds = [];

            foreach ($results as $result) {
                if (in_array('product', $result->type)) {
                    if (is_array($result->parentid) && count($result->parentid) > 0) {
                        $productIds[] = ProductIdStrategy::ID_OFFSET + $result->parentid[0];
                    } else {
                        $productIds[] = absint($result->id) - 1e7 + ProductIdStrategy::ID_OFFSET;
                    }
                }
            }

            $productIds = array_map('intval', array_unique($productIds));
            $query->query_vars['found_posts'] = count($productIds);

            if (count($productIds) > 0) {
                $pages = array_chunk($productIds, $query->query_vars['posts_per_page']);
                $productIds = $pages[$page - 1];

                $query->set('post__in', $productIds);
            }
        }

        return $query;
    }

    /**
     * @param $posts
     * @param WP_Query $query
     * @return WP_Post[]|null
     */
    public function get_posts($posts, WP_Query $query)
    {
        // netko je već vratio postove
        if (is_array($posts) && !empty($posts)) {
            return $posts;
        }

        if (!is_admin() && $query->is_main_query() && $query->is_search()) {
            $query = $this->product_search($query);
        }

        // ne tražimo proizvode uopće
        if (!$query->is_search() && (@$query->query_vars['wc_query'] ?? '') !== 'product_query' && $query->query_vars['post_type'] !== 'product') {
            return $posts;
        }

        switch ($query->query_vars['orderby'] ?? '') {
            case 'stock':
                $sort = ProductRepository::SORT_STOCK_DESC;
                break;
            case 'menu_order title':
                $sort = ProductRepository::SORT_NAME_ASC;
                break;
            case 'price':
                $sort = $query->query_vars['order'] === 'DESC'
                    ? ProductRepository::SORT_PRICE_DESC
                    : ProductRepository::SORT_PRICE_ASC;
                break;
            case 'date ID':
            default:
                $sort = ProductRepository::SORT_DATE_DESC;
                break;
        }

        $limit = $query->query_vars['posts_per_page'];
        $page = max(1, intval($query->query_vars['paged']));
        $offset = ($page - 1) * $limit;

        switch (true) {
            case isset($query->query_vars['post__in']) && count($query->query_vars['post__in']) > 0:
                $productIds = array_map('intval', $query->query_vars['post__in']);
                foreach ($productIds as $productId) {
                    if ($productId < ProductIdStrategy::ID_OFFSET) {
                        return $posts;
                    }
                }
                $products = $this->repo->getByIds($productIds);
                break;
            case preg_match('/-(\d{7})$/', $query->query_vars['name'], $matches):
                $productId = intval($matches[1]);
                $products = [$this->repo->getById($productId)];
                break;
            case ($query->query_vars['post_status'] ?? '') === 'featured':
                $products = $this->repo->getFeaturedProducts();
                break;
            default:
                foreach ($query->tax_query->queries ?? [] as $taxQuery) {
                    if (!is_array($taxQuery)) {
                        continue;
                    }

                    if (preg_match('/^pa_(.+)$/', $taxQuery['taxonomy'], $matches)) {
                        $attributeQuery[$matches[1]] = $taxQuery['terms'];
                    } else if ($taxQuery['taxonomy'] === 'product_cat') {
                        $categorySlug = $taxQuery['terms'][0] ?? '';
                    }
                }

                $category = isset($categorySlug) ? get_term_by('slug', $query->query_vars['product_cat'], 'product_cat') : null;

                if (isset($attributeQuery)) {
                    $products = $this->repo->getByAttributes($attributeQuery, $category ? $category->term_id : null, $limit, $offset,$sort);
                } else if (isset($category)) {
                    $products = $this->repo->getByCategorySlug($category->slug, $limit, $offset, $sort);
                }
                break;
        }

        if (!isset($products)) {
            return $posts;
        }

        $found_posts = $query->query_vars['found_posts'] ?? $this->repo->getTotalProducts();

        $query->set('no_found_rows', $this->repo->getTotalProducts());
        $query->found_posts = $found_posts;
        $query->max_num_pages = ceil($found_posts / $limit);
        $query->thumbnails_cached = true;

        $posts = array_map([$this, 'load_post'], $products);
        return apply_filters('the_posts', $posts, $query);
    }

    /**
     * @param int $objectId
     * @return array
     */
    protected function get_metadata_by_id(int $objectId): array
    {
        if (!($this->products[$objectId] ?? null) instanceof Product) {
            $this->products[$objectId] = $this->repo->getById($objectId);
        }

        $images = $this->products[$objectId]->getImages();
        $image_ids = array_map([$this->images, 'getAttachmentIdByGathPath'], $images);
        $image_ids = array_filter($image_ids);

        return [
            '_sku'                   => $this->products[$objectId]->getSku(),
            'total_sales'            => '',
            '_tax_status'            => '',
            '_tax_class'             => '',
            '_manage_stock'          => 'yes',
            '_low_stock_amount'      => '',
            '_sold_individually'     => 'no',
            '_weight'                => '',
            '_length'                => '',
            '_width'                 => '',
            '_height'                => '',
            '_upsell_ids'            => '',
            '_crosssell_ids'         => '',
            '_purchase_note'         => '',
            '_default_attributes'    => '',
            '_virtual'               => 'no',
            '_downloadable'          => 'no',
            '_download_limit'        => '',
            '_download_expiry'       => '',
            '_thumbnail_id'          => count($image_ids) > 0 ? $image_ids[0] : '',
            '_wc_average_rating'     => '',
            '_wc_rating_count'       => '',
            '_wc_review_count'       => '',
            '_product_image_gallery' => count($image_ids) > 1 ? implode(',', array_slice($image_ids, 1)) : '',
        ];
    }

    /**
     * @param Product $product
     * @return WP_Post
     */
    public function load_post(Product $product): WP_Post
    {
        $postDate = preg_replace('/\.\d+$/', '', $product->getDateModified()->format('Y-m-d H:i:s'));
        $postTitle = $product->getName();

        $postName = sprintf('%s-%d', sanitize_title($product->getName()), $product->getId());
        $description = preg_replace('/<([a-z]+)[^>]+>/', '<\1>', $product->getDescription());

        $post = new WP_Post((object) [
            'ID' => $product->getId(),
            'post_author' => 1,
            'post_date' => $postDate,
            'post_date_gmt', $postDate,
            'post_content' => $description,
            'post_title' => $postTitle,
            'post_excerpt' => $product->getShortDescription(),
            'post_status' => 'publish',
            'comment_status' => 'closed',
            'ping_status' => 'closed',
            'post_password' => '',
            'post_name' => $postName,
            'to_ping' => '',
            'pinged' => '',
            'post_modified' => $postDate,
            'post_modified_gmt' => $postDate,
            'post_content_filtered' => '',
            'post_parent' => 0,
            'guid' => sprintf('%s?p=%d', get_option('siteurl'), $product->getId()),
            'post_type' => 'product',
            'filter' => 'raw',
        ]);

        wp_cache_add($post->ID, $post, 'posts');
        $this->products[$product->getId()] = $product;

        return $post;
    }

    public function setup_post(WP_Post $post)
    {
        $GLOBALS['post'] = $post;
    }

    public function wc_add_to_cart_product_id($productId)
    {
        if ($productId > IdStrategy::ID_OFFSET) {
            get_posts([
                'post_type' => 'product',
                'include' => [$productId],
            ]);
        }

        return $productId;
    }

    public function preload_cart_products()
    {
        $session = WC()->session;

        if (is_object($session) && method_exists($session, 'get')) {
            $cart = $session->get('cart');
            $postIds = [];

            foreach ($cart ?? [] as $key => $value) {
                if (isset($value['product_id']) && $value['product_id'] > IdStrategy::ID_OFFSET) {
                    $postIds[] = $value['product_id'];
                }
            }

            if (!empty($postIds)) {
                get_posts([
                    'post_type' => 'product',
                    'include' => $postIds,
                ]);
            }
        }
    }

    /** @noinspection PhpUnusedPrivateMethodInspection */
    private function image2postId(string $gathPath)
    {
        global $wpdb;
        $sql = "SELECT post_id FROM wp_postmeta WHERE meta_key = '_gath_path' AND meta_value = '{$gathPath}'";
        $attachmentId = $wpdb->get_col($sql);

        return count($attachmentId) > 0 ? intval($attachmentId[0]) : 'gath-' . $gathPath;
    }
}
